#ifndef Map_H
#define	Map_H

#include <Arduino.h>

template <class K, class T>
class MapItem {
private:
    K keyVal;
    T itemVal;
    MapItem<K, T> * next;

public:

    MapItem(K key, T val) {
        keyVal = key;
        itemVal = val;
        next = NULL;
    }

    K getKey() {
        return keyVal;
    }
    
    void setVal(T val) {
        itemVal = val;
    }

    T getVal() {
        return itemVal;
    }

    void setNext(MapItem<K, T> * n) {
        next = n;
    }

    MapItem<K, T> * getNext() {
        return next;
    }
};


template <class K, class T>
class Map {
private:
    MapItem<K, T> * head;
    int count;

public:
    Map() {
        head = NULL;
        count = 0;
    }

    ~Map() {
        clear();
    }
    
    void clear() {
        MapItem<K, T> * it = head;

        while (it != NULL) {
            head = it->getNext();
            delete it;
            it = head;
        }

        head = NULL;
        count = 0;
    }
    
    void add(K key, T val) {
        if (!containsKey(key)) {
            MapItem<K, T> * item = new MapItem<K, T>(key, val);
            item->setNext(head);
            head = item;
            count++;
        }
    }
    
    void addOrReplace(K key, T val) {
        if (!containsKey(key)) {
            add(key, val);
        } else {
            MapItem<K, T> * it = head;

            while (it != NULL) {
                if (it->getKey() == key) {
                    it->setVal(val);
                    return;
                }

                it = it->getNext();
            }
        }
    }
    
    bool contains(T val) {
        MapItem<K, T> * it = head;
        
        while (it != NULL) {
            if (it->getVal() == val) {
                return true;
            }
            
            it = it->getNext();
        }
        
        return false;
    }
    
    bool containsKey(K key) {
        MapItem<K, T> * it = head;
        
        while (it != NULL) {
            if (it->getKey() == key) {
                return true;
            }
            
            it = it->getNext();
        }
        
        return false;
    }
    
    void remove(K key) {
        MapItem<K, T> * it = head;
        MapItem<K, T> * prevIt = NULL;
        
        while (it != NULL) {
            if (it->getKey() == key) {
                if (prevIt != NULL) {
                    prevIt->setNext(it->getNext());
                } else {
                    head = it->getNext();
                }
                delete it;
                count--;
                return;
            }

            prevIt = it;
            it = it->getNext();
        }
        
    }
    
    bool get(K key, T & val) {
        MapItem<K, T> * it = head;
        
        while (it != NULL) {
            if (it->getKey() == key) {
                val = it->getVal();
                return true;
            }
            
            it = it->getNext();
        }
        
        return false;
    }

    bool isEmpty() {
        return count == 0;
    }
    
    int size() {
        return count;
    }
    
    void forAll(void f(K k, T v)) {
        MapItem<K, T> * it = head;
        while (it != NULL) {
            f(it->getKey(), it->getVal());
            it = it->getNext();
        }
    }

    void forAll(void f(K k, T v, void * data), void * data) {
        MapItem<K, T> * it = head;
        while (it != NULL) {
            f(it->getKey(), it->getVal(), data);
            it = it->getNext();
        }
    }
    
    void copyFrom(Map<K, T> * otherMap) {
        MapItem<K, T> * it = otherMap->head;
        while (it != NULL) {
            add(it->getKey(), it->getVal());
            it = it->getNext();
        }
    }
    
    void moveFrom(Map<K, T> * otherMap) {
        copyFrom(otherMap);
        otherMap->clear();
    }
    
    void copyReplaceFrom(Map<K, T> * otherMap) {
        MapItem<K, T> * it = otherMap->head;
        while (it != NULL) {
            addOrReplace(it->getKey(), it->getVal());
            it = it->getNext();
        }
    }
    
    void moveReplaceFrom(Map<K, T> * otherMap) {
        copyReplaceFrom(otherMap);
        otherMap->clear();
    }
    
};

#endif

